#include "CDepthsorting.h"
#include <iostream>
#include <time.h>

CDepthsorting::CDepthsorting(){

}

CDepthsorting::~CDepthsorting(){
	cleanUp();
}

int CDepthsorting::addNode(int x, int w, int y, int d, int z, int h, bool dynamic, int responce){
	//Create a new node

	node* newNode = new node;

	//Assign the desired values to the node
	newNode->x = x;
	newNode->w = w;
	newNode->y = y;
	newNode->d = d;
	newNode->z = z;
	newNode->h = h;
	newNode->dynamic = dynamic;
	newNode->intersectionResponce = responce;

	//Store this newly created node in the node index
	nodeIndex.push_back(newNode);

	edgeList.clear();
	storedEdgeList.clear();

	//Return the size of the node index. This will also be the id of the node
	return nodeIndex.size() - 1;
}

int CDepthsorting::moveNode(int nodeId, int x, int w, int y, int d, int z, int h){

	if ((unsigned int)nodeId < nodeIndex.size()){

		node* currentNode = nodeIndex.at(nodeId);
		if (currentNode->dynamic == true){
			currentNode->x = x;
			currentNode->w = w;
			currentNode->y = y;
			currentNode->d = d;
			currentNode->z = z;
			currentNode->h = h;
			return 1;
		}
	}
	return 0;
}

void CDepthsorting::determineEdges(bool flag){
	/*This function determines for every pair of objects wheter one has to
	be drawn in front of the other or if it does not matter. If it does matter
	this relation (edge) is stored in edgeList. There are two different
	procedures. One for the first time, when every object has to be tested
	against every other object, and one for when that has been done before
	and only the dynamic objects have to be checked against everything.
	When two objects overlap and one or both of them have a special way
	to handle 3d overlap (as opposed to the 2d hexagon overlap), the code
	checks wheter the two cuboids intersect. If they do, the relation
	calculated based on the hexagonal 2d projection overlap may be changed.*/

	if (flag == ALL){
		//Save the size of the node index
		size_t nodeIndexSize = nodeIndex.size();

		//Iterate through the node index
		for (unsigned int i = 0; i < nodeIndexSize; i++){
			//Get the node from the node index
			node* nodeA = nodeIndex.at(i);
			//Store the responce of a to intersections for later use
			int responceA = nodeA->intersectionResponce;
			nodeA->takenOut = false;

			//Iterate through the rest of the list
			for (unsigned int j = i + 1; j < nodeIndexSize; j++){
				//Get the second node from the node index
				node* nodeB = nodeIndex.at(j);
				//Calculate which one is behind
				int result = hexagonOverlap(*nodeA, *nodeB);
				//If the nodes turn out to overlap...
				if (result != NO_OVERLAP){
					//Handle the  cases in which the objects overlap in 3d
					int responceB = nodeB->intersectionResponce;
					//Check for intersections only if an intersection would make a difference
					if (responceA != R_NONE || responceB != R_NONE)
						result = handleIntersection(nodeA, nodeB, responceA, responceB, result);
					//Store the edge that was found//
					if (result == A_IS_BEHIND_B){
						edge newEdge = { i, j };
						edgeList.push_back(newEdge);
						nodeB->numberOfIncommingEdges++;
					}
					else if (result == B_IS_BEHIND_A){
						edge newEdge = { j, i };
						edgeList.push_back(newEdge);
						nodeA->numberOfIncommingEdges++;
					}
				}
			}
		}
		//Copy the calculated edges to the stored edge list



		//Remove the dynamic edges out of the stored edge list
		//This list now has just the static edges and can be used
		//for all future sorts

		storedEdgeList = edgeList;

		removeDynamicEdges();

	}
	else if (flag == DYNAMIC){
		edgeList = storedEdgeList;
		for (std::list<edge>::iterator iterator = edgeList.begin(); iterator != edgeList.end(); iterator++){
			nodeIndex.at((*iterator).to)->numberOfIncommingEdges++;
		}
		//Save the size of the node index
		size_t nodeIndexSize = nodeIndex.size();
		//Iterate through the node index
		for (unsigned int i = 0; i < nodeIndexSize; i++){
			//Get the node from the node index
			node* nodeA = nodeIndex.at(i);
			//Store the responce of a to intersections for later use
			int responceA = nodeA->intersectionResponce;
			//Mark the node as not yet taken out
			nodeA->takenOut = false;
			bool ADynamic = nodeA->dynamic;
			if (ADynamic == true){
				//Iterate through the rest of the list
				for (unsigned int j = 0; j < nodeIndexSize; j++){
					//Get the second node from the node index
					node* nodeB = nodeIndex.at(j);
					bool BDynamic = nodeB->dynamic;
					//If node b is not the same as node a or dynamic and earlier in the list than a (in
					//which case a and b have already been checked agains eachother)
					if ((j != i) && !(j < i && BDynamic == true)){
						//Calculate which one is behind
						int result = hexagonOverlap(*nodeA, *nodeB);
						//If the nodes turn out to overlap...
						if (result != NO_OVERLAP){
							//Handle the  cases in which the objects overlap in 3d
							int responceB = nodeB->intersectionResponce;
							//Check for intersections only if an intersection would make a difference
							if (responceA != R_NONE || responceB != R_NONE)
								result = handleIntersection(nodeA, nodeB, responceA, responceB, result);
							//Store the edge that was found//
							if (result == A_IS_BEHIND_B){
								edge newEdge = { i, j };
								edgeList.push_back(newEdge);
								nodeB->numberOfIncommingEdges++;
							}
							else if (result == B_IS_BEHIND_A){
								edge newEdge = { j, i };
								edgeList.push_back(newEdge);
								nodeA->numberOfIncommingEdges++;
							}
						}
					}
				}
			}
		}
	}

}

int CDepthsorting::handleIntersection(node* a, node* b, int responceA, int responceB, int result){
	/*This function adjusts the result of the hexagon intersect
	in the cases that the objects intersect. This way one can set
	how an intersection should be handled for every individual object.
	*/

	//First calculate if there is an intersection
	int c1[3];
	int c2[3];
	int c3[3];
	int c4[3];

	c1[0] = a->x;
	c2[0] = c1[0] + a->w;
	c1[1] = a->y;
	c2[1] = c1[1] + a->d;
	c1[2] = a->z;
	c2[2] = c1[2] + a->h;

	c3[0] = b->x;
	c4[0] = c3[0] + b->w;
	c3[1] = b->y;
	c4[1] = c3[1] + b->d;
	c3[2] = b->z;
	c4[2] = c3[2] + b->h;

	bool dimensionWithoutOverlapFound = false;
	//Check for every dimension
	for (int dim = 0; dim < 3; dim++){
		//If there is no overlap
		if (!(c1[dim] < c4[dim] && c3[dim] < c2[dim]))
			dimensionWithoutOverlapFound = true;
	}
	//If there is no intersection occuring
	if (dimensionWithoutOverlapFound)
		//Do not alter the result
		return result;

	int responce = (responceA != R_NONE ? responceA : responceB);

	switch (responce){
	case R_NONE:
		return result;
		break;

	case R_PLACE_BEHIND:
		if (responceA == R_PLACE_BEHIND)
			return A_IS_BEHIND_B;
		else
			return B_IS_BEHIND_A;
		break;

	case R_PLACE_IN_FRONT:
		if (responceA == R_PLACE_IN_FRONT)
			return B_IS_BEHIND_A;
		else
			return A_IS_BEHIND_B;
		break;

	case R_CHECK_X:
	case R_CHECK_Y:
	case R_CHECK_Z:
		int dim = ((responce == R_CHECK_Y) + 2 * (responce == R_CHECK_Z));
		
		if (0.5*(c1[dim] + c2[dim]) < 0.5*(c3[dim] + c4[dim]))
			return A_IS_BEHIND_B;
		else
			return B_IS_BEHIND_A;
		break;

	}

	return true;
}

int CDepthsorting::hexagonOverlap(node a, node b){
	int Xa = a.x;
	int Ya = a.y;
	int Za = a.z;
	int XA = Xa + a.w;
	int YA = Ya + a.d;
	int ZA = Za + a.h;

	int Xb = b.x;
	int Yb = b.y;
	int Zb = b.z;
	int XB = Xb + b.w;
	int YB = Yb + b.d;
	int ZB = Zb + b.h;

	int A = Xa - YA;
	int B = XA - Ya;
	int C = Xb - YB;
	int D = XB - Yb;

	//Check for overlap
	if ((Xa - YA < XB - Yb) && (Xb - YB < XA - Ya))
		if ((Xa - ZA <= XB - Zb) && (Xb - ZB <= XA - Za))
			if ((Za - YA <= ZB - Yb) && (Zb - YB <= ZA - Ya))
			{
				//Find out which one is in front
				if (XA <= Xb || YA <= Yb || ZA <= Zb)
					return A_IS_BEHIND_B;
				else
					return B_IS_BEHIND_A;
			}

	return NO_OVERLAP;
}

void CDepthsorting::topologicalSort(){

	resultList.clear();

	if (storedEdgeList.empty())
		//Calculate the edges for all nodes
		determineEdges(ALL);
	else
		//Calculate the edges for the dynamic nodes
		determineEdges(DYNAMIC);


	//for (unsigned int i = 0; i < nodeIndex.size(); i++)
	//Now the sorting will be done//
	//Keep sorting until every node is in the resulting queue
	while (resultList.size() != nodeIndex.size()){

		//First time sort, from scratch//
		//Find and mark all nodes without incomming edges//
		//Iterate through the node index
		for (unsigned int i = 0; i < nodeIndex.size(); i++){

			//Prevent having to look up the node twice
			node* currentNode = nodeIndex.at(i);

			//If the node has no incomming edges...
			if (currentNode->numberOfIncommingEdges == 0 && currentNode->takenOut == false){
				//Mark it to indicate that it has been taken out
				currentNode->takenOut = true;
				//Remove Edges
				//Chain removing is enabled here
				removeEdgesFrom(i);
			}

		}

		if (!edgeList.empty())
			attemptLoopRemoval();
	}



}

void CDepthsorting::removeDynamicEdges(){
	for (std::list<edge>::iterator iterator = storedEdgeList.begin(); iterator != storedEdgeList.end();){
		edge currentEdge = (*iterator);
		node* currentNode = nodeIndex.at(currentEdge.from);
		if (currentNode->dynamic == true){
			iterator = storedEdgeList.erase(iterator);
		}
		else{
			currentNode = nodeIndex.at(currentEdge.to);
			if (currentNode->dynamic == true){
				iterator = storedEdgeList.erase(iterator);
			}
			else
				iterator++;
		}
	}
}

void CDepthsorting::attemptLoopRemoval(){

	//Attempt too remove loops//
	//Make an iterator and store it to avoid having to do it twice
	std::list<edge>::iterator iterator = edgeList.begin();
	//If the edge list is not empty...
	if (!edgeList.empty()){
		//Get the edge that is the first element in the list
		edge currentEdge = (*iterator);
		//Tell the node this edge is going to that it has lost an edge
		nodeIndex.at(currentEdge.to)->numberOfIncommingEdges--;
		//Remove the edge from the list
		iterator = edgeList.erase(iterator);
	}
}

void CDepthsorting::removeEdgesFrom(int nodeId){
	//Take out this node//
	//Get the node from the node index
	node* currentNode = nodeIndex.at(nodeId);
	//Mark this node as taken out
	currentNode->takenOut = true;
	//Place the node in the results queue
	resultList.push_back(nodeId);

	//Iterate through the entire edge list and remove any edges from the node with nodeId//
	//Iterate through the edge list
	for (std::list<edge>::iterator iterator = edgeList.begin(); iterator != edgeList.end();){
		//Prevent having to look up the edge twice
		edge currentEdge = (*iterator);
		//If the current node is comming from the node we are removing edges from...
		if (currentEdge.from == nodeId){
			//First remove this edge from the list and store the location of the next element
			iterator = edgeList.erase(iterator);
			//Prevent having to look up the node twice
			node* otherNode = nodeIndex.at(currentEdge.to);
			//Subtract one from the number of incomming edges the other node has
			otherNode->numberOfIncommingEdges--;
			//If this other node now has no more edges...
			if (otherNode->numberOfIncommingEdges == 0){
				//Call this same function for that node
				removeEdgesFrom(currentEdge.to);
				//Since it is unknown how many elements have been removed, the iterator has to be set back to the start
				iterator = edgeList.begin();
			}

		}
		else//If the current node is not the one we are removing edges from
			//Simply jump to the next one for the next cycle of the loop
			iterator++;
	}

}

void CDepthsorting::optimizeStoredEdgeList(){

	for (std::list<edge>::iterator iterator = storedEdgeList.begin(); iterator != storedEdgeList.end();){
		edge currentEdge = { (*iterator).from, (*iterator).to };
		if (checkObsolete(currentEdge.from, currentEdge.from, currentEdge.to, 2)){
			iterator = storedEdgeList.erase(iterator);
		}
		else
			iterator++;
	}
}

bool CDepthsorting::checkObsolete(int startNode, int currentNode, int targetNode, int lifeTime){

	if (lifeTime == 0)
		return false;

	edge currentEdge;
	for (std::list<edge>::iterator iterator = edgeList.begin(); iterator != edgeList.end(); iterator++){

		currentEdge = (*iterator);
		if (!(currentEdge.from == startNode && currentEdge.to == targetNode)){
			if (currentEdge.from == currentNode){
				if (currentEdge.to == targetNode)
					return true;
				else if (checkObsolete(startNode, currentEdge.to, targetNode, lifeTime - 1))
					return true;

			}
		}
	}
	return false;
}

void CDepthsorting::cleanUp(){
	//Delete all nodes in the node index//

	//Iterate through all nodes
	for (unsigned int i = 0; i < nodeIndex.size(); i++){

		//Store the current node to prevent having to look it up twice
		node* currentNode = nodeIndex.at(i);

		//If the node is not empty it is destroyed
		if (currentNode != nullptr)
			delete currentNode;
	}
}

int CDepthsorting::nodeIndexGetSize(){
	return (int)nodeIndex.size();
}

void CDepthsorting::copyResultToString(char* result){
	int i = 0;
	for (std::list<int>::iterator resultIterator = resultList.begin(); resultIterator != resultList.end(); resultIterator++, i += 2){
		int id = (*resultIterator);
		result[i] = (id % 127) + 1;
		result[i + 1] = (id / 127) + 1;
	}
}